package proof

import proof.types.trees.BinaryTrees.{*, given}
import proof.types.formulas.*
import proof.types.formulas.ProofSystem.*

object InOrderMirror:
  object Variables:
    // T is a preset type
    val x = Variable[T]("x")
    val xs = ListVariable[T]("xs")
    val ys = ListVariable[T]("ys")
    val zs = ListVariable[T]("zs")
    val tree = BinaryTreeVariable[T]("tree")
    val left = BinaryTreeVariable[T]("left")
    val right = BinaryTreeVariable[T]("right")
  import Variables.*

  object Axioms:
    object InOrder:
      val EmptyCase = Axiom:
        Empty.inOrder === Nil
      val NodeCase = Axiom:
        Node(left, x, right).inOrder ===
        left.inOrder ++ ((x :: Nil) ++ right.inOrder)

    object Mirror:
      val EmptyCase = Axiom:
        Empty.mirror === Empty
      val NodeCase = Axiom:
        Node(left, x, right).mirror ===
        Node(right.mirror, x, left.mirror)

    object Reverse:
      val NilCase = Axiom:
        Nil.reverse === Nil
      val ConsCase = Axiom:
        (x :: xs).reverse === xs.reverse ++ (x :: Nil)
      val AppendCase = Axiom:
        (xs ++ ys).reverse === ys.reverse ++ xs.reverse
      val SingleCase = Axiom:
        (x :: Nil).reverse === x :: Nil

    object Append:
      val NilCase = Axiom:
        (Nil ++ xs) === xs
      val ConsCase = Axiom:
        (x :: xs) ++ ys === x :: (xs ++ ys)
      val Associativity = Axiom:
        (xs ++ ys) ++ zs === xs ++ (ys ++ zs)

  object InductionHypotheses:
    val IHLeft = left.mirror.inOrder === left.inOrder.reverse
    val IHRight = right.mirror.inOrder === right.inOrder.reverse
  import InductionHypotheses.*

  final val EmptyCaseStatement =
    Empty.mirror.inOrder === Empty.inOrder.reverse

  final val NodeCaseStatement =
    IHLeft ==> (IHRight ==>
      (Node(left, x, right).mirror.inOrder === Node(left, x, right).inOrder.reverse))

trait InOrderMirrorTheorems:
  import InOrderMirror.*
  import Variables.*
  import Axioms.*
  import InductionHypotheses.*

  val EmptyCase: Theorem[Expr[T]]
  val NodeCase: Theorem[Expr[T]]
