package proof

import proof.types.formulas.ProofSystem.*
import proof.types.formulas.*
import proof.InOrderMirror
import proof.types.trees.BinaryTrees.{*, given}

/** Since the proofs are already checked by the proof system, the tests simply
  * verify that the theorems have been run and that their statements and axioms
  * have not been altered.
  */
class ProofTest extends munit.FunSuite:
  import InOrderMirror.*
  val allAxiomStatements = Seq(
    Axioms.InOrder.EmptyCase,
    Axioms.InOrder.NodeCase,
    Axioms.Mirror.EmptyCase,
    Axioms.Mirror.NodeCase,
    Axioms.Reverse.NilCase,
    Axioms.Reverse.ConsCase,
    Axioms.Reverse.AppendCase,
    Axioms.Reverse.SingleCase,
    Axioms.Append.NilCase,
    Axioms.Append.ConsCase,
    Axioms.Append.Associativity
  )

  def checkStatement[T](theorem: Theorem[T], statement: Formula[T]) =
    assertEquals(
      theorem.statement,
      statement,
      "The theorem statement has been altered."
    )

  def checkAssumptions[T, T1 >: T](
      theorem: Theorem[T],
      justifications: Seq[Justification[T1]]
  ) =
    val extraneousJustifications = theorem.innerProof.steps.collect {
      case Equivalent(_, justification) => justification
      case Equal(_, justification)      => justification
    }.filterNot(justifications.contains)

    assert(
      extraneousJustifications.isEmpty,
      f"A non-predefined axiom was used: ${extraneousJustifications.head}."
    )

  test("Base Case (Example) Passes (0pts)"):
    // check that the theorem itself is intitialized and checked
    import InOrderMirrorProof.*
    EmptyCase.verify
    checkStatement(EmptyCase, InOrderMirror.EmptyCaseStatement)
    checkAssumptions(EmptyCase, allAxiomStatements)

  test("Inductive Case Passes (40pts)"):
    import InOrderMirrorProof.*
    NodeCase.verify
    checkStatement(NodeCase, InOrderMirror.NodeCaseStatement)
    checkAssumptions(
      NodeCase,
      allAxiomStatements
        :+ Assumption(InOrderMirror.InductionHypotheses.IHLeft)
        :+ Assumption(InOrderMirror.InductionHypotheses.IHRight)
    )
